@@ -1,7 +1,7 @@
 # Authors: David Goodger, Dethe Elza
 # Contact: goodger@users.sourceforge.net
-# Revision: $Revision: 1.24 $
-# Date: $Date: 2004/07/26 22:32:24 $
+# Revision: $Revision: 4229 $
+# Date: $Date: 2005-12-23 00:46:16 +0100 (Fri, 23 Dec 2005) $
 # Copyright: This module has been placed in the public domain.
 
 """Miscellaneous directives."""
@@ -11,6 +11,7 @@ __docformat__ = 'reStructuredText'
 import sys
 import os.path
 import re
+import time
 from docutils import io, nodes, statemachine, utils
 from docutils.parsers.rst import directives, roles, states
 from docutils.transforms import misc
@@ -21,23 +22,31 @@ except ImportError:
     urllib2 = None
 
 
+standard_include_path = os.path.join(os.path.dirname(states.__file__),
+                                     'include')
+
 def include(name, arguments, options, content, lineno,
             content_offset, block_text, state, state_machine):
     """Include a reST file as part of the content of this reST file."""
+    if not state.document.settings.file_insertion_enabled:
+        warning = state_machine.reporter.warning(
+              '"%s" directive disabled.' % name,
+              nodes.literal_block(block_text, block_text), line=lineno)
+        return [warning]
     source = state_machine.input_lines.source(
         lineno - state_machine.input_offset - 1)
     source_dir = os.path.dirname(os.path.abspath(source))
-    path = ''.join(arguments[0].splitlines())
-    if path.find(' ') != -1:
-        error = state_machine.reporter.error(
-              '"%s" directive path contains whitespace.' % name,
-              nodes.literal_block(block_text, block_text), line=lineno)
-        return [error]
+    path = directives.path(arguments[0])
+    if path.startswith('<') and  path.endswith('>'):
+        path = os.path.join(standard_include_path, path[1:-1])
     path = os.path.normpath(os.path.join(source_dir, path))
     path = utils.relative_path(None, path)
+    encoding = options.get('encoding', state.document.settings.input_encoding)
     try:
+        state.document.settings.record_dependencies.add(path)
         include_file = io.FileInput(
-            source_path=path, encoding=state.document.settings.input_encoding,
+            source_path=path, encoding=encoding,
+            error_handler=state.document.settings.input_encoding_error_handler,
             handle_io_errors=None)
     except IOError, error:
         severe = state_machine.reporter.severe(
@@ -45,7 +54,14 @@ def include(name, arguments, options, content, lineno,
               % (name, error.__class__.__name__, error),
               nodes.literal_block(block_text, block_text), line=lineno)
         return [severe]
-    include_text = include_file.read()
+    try:
+        include_text = include_file.read()
+    except UnicodeError, error:
+        severe = state_machine.reporter.severe(
+              'Problem with "%s" directive:\n%s: %s'
+              % (name, error.__class__.__name__, error),
+              nodes.literal_block(block_text, block_text), line=lineno)
+        return [severe]
     if options.has_key('literal'):
         literal_block = nodes.literal_block(include_text, include_text,
                                             source=path)
@@ -58,7 +74,8 @@ def include(name, arguments, options, content, lineno,
         return []
 
 include.arguments = (1, 0, 1)
-include.options = {'literal': directives.flag}
+include.options = {'literal': directives.flag,
+                   'encoding': directives.encoding}
 
 def raw(name, arguments, options, content, lineno,
         content_offset, block_text, state, state_machine):
@@ -70,7 +87,15 @@ def raw(name, arguments, options, content, lineno,
     Content may be included inline (content section of directive) or
     imported from a file or url.
     """
-    attributes = {'format': arguments[0]}
+    if ( not state.document.settings.raw_enabled
+         or (not state.document.settings.file_insertion_enabled
+             and (options.has_key('file') or options.has_key('url'))) ):
+        warning = state_machine.reporter.warning(
+              '"%s" directive disabled.' % name,
+              nodes.literal_block(block_text, block_text), line=lineno)
+        return [warning]
+    attributes = {'format': ' '.join(arguments[0].lower().split())}
+    encoding = options.get('encoding', state.document.settings.input_encoding)
     if content:
         if options.has_key('file') or options.has_key('url'):
             error = state_machine.reporter.error(
@@ -91,14 +116,24 @@ def raw(name, arguments, options, content, lineno,
         path = os.path.normpath(os.path.join(source_dir, options['file']))
         path = utils.relative_path(None, path)
         try:
-            raw_file = open(path)
+            state.document.settings.record_dependencies.add(path)
+            raw_file = io.FileInput(
+                source_path=path, encoding=encoding,
+                error_handler=state.document.settings.input_encoding_error_handler,
+                handle_io_errors=None)
         except IOError, error:
             severe = state_machine.reporter.severe(
                   'Problems with "%s" directive path:\n%s.' % (name, error),
                   nodes.literal_block(block_text, block_text), line=lineno)
             return [severe]
-        text = raw_file.read()
-        raw_file.close()
+        try:
+            text = raw_file.read()
+        except UnicodeError, error:
+            severe = state_machine.reporter.severe(
+                  'Problem with "%s" directive:\n%s: %s'
+                  % (name, error.__class__.__name__, error),
+                  nodes.literal_block(block_text, block_text), line=lineno)
+            return [severe]
         attributes['source'] = path
     elif options.has_key('url'):
         if not urllib2:
@@ -108,17 +143,27 @@ def raw(name, arguments, options, content, lineno,
                   '"urllib2" module).' % name,
                   nodes.literal_block(block_text, block_text), line=lineno)
             return [severe]
+        source = options['url']
         try:
-            raw_file = urllib2.urlopen(options['url'])
+            raw_text = urllib2.urlopen(source).read()
         except (urllib2.URLError, IOError, OSError), error:
             severe = state_machine.reporter.severe(
                   'Problems with "%s" directive URL "%s":\n%s.'
                   % (name, options['url'], error),
                   nodes.literal_block(block_text, block_text), line=lineno)
             return [severe]
-        text = raw_file.read()
-        raw_file.close()
-        attributes['source'] = options['file']
+        raw_file = io.StringInput(
+            source=raw_text, source_path=source, encoding=encoding,
+            error_handler=state.document.settings.input_encoding_error_handler)
+        try:
+            text = raw_file.read()
+        except UnicodeError, error:
+            severe = state_machine.reporter.severe(
+                  'Problem with "%s" directive:\n%s: %s'
+                  % (name, error.__class__.__name__, error),
+                  nodes.literal_block(block_text, block_text), line=lineno)
+            return [severe]
+        attributes['source'] = source
     else:
         error = state_machine.reporter.warning(
             'The "%s" directive requires content; none supplied.' % (name),
@@ -129,7 +174,8 @@ def raw(name, arguments, options, content, lineno,
 
 raw.arguments = (1, 0, 1)
 raw.options = {'file': directives.path,
-               'url': directives.path}
+               'url': directives.uri,
+               'encoding': directives.encoding}
 raw.content = 1
 
 def replace(name, arguments, options, content, lineno,
@@ -148,8 +194,7 @@ def replace(name, arguments, options, content, lineno,
             messages = []
             for node in element:
                 if isinstance(node, nodes.system_message):
-                    if node.has_key('backrefs'):
-                        del node['backrefs']
+                    node['backrefs'] = []
                     messages.append(node)
             error = state_machine.reporter.error(
                 'Error in "%s" directive: may contain a single paragraph '
@@ -167,7 +212,7 @@ def replace(name, arguments, options, content, lineno,
 replace.content = 1
 
 def unicode_directive(name, arguments, options, content, lineno,
-                         content_offset, block_text, state, state_machine):
+                      content_offset, block_text, state, state_machine):
     r"""
     Convert Unicode character codes (numbers) to characters.  Codes may be
     decimal numbers, hexadecimal numbers (prefixed by ``0x``, ``x``, ``\x``,
@@ -181,6 +226,14 @@ def unicode_directive(name, arguments, options, content, lineno,
             'substitution definition.' % (name),
             nodes.literal_block(block_text, block_text), line=lineno)
         return [error]
+    substitution_definition = state_machine.node
+    if options.has_key('trim'):
+        substitution_definition.attributes['ltrim'] = 1
+        substitution_definition.attributes['rtrim'] = 1
+    if options.has_key('ltrim'):
+        substitution_definition.attributes['ltrim'] = 1
+    if options.has_key('rtrim'):
+        substitution_definition.attributes['rtrim'] = 1
     codes = unicode_comment_pattern.split(arguments[0])[0].split()
     element = nodes.Element()
     for code in codes:
@@ -196,26 +249,42 @@ def unicode_directive(name, arguments, options, content, lineno,
     return element.children
 
 unicode_directive.arguments = (1, 0, 1)
+unicode_directive.options = {'trim': directives.flag,
+                             'ltrim': directives.flag,
+                             'rtrim': directives.flag}
 unicode_comment_pattern = re.compile(r'( |\n|^)\.\. ')
 
 def class_directive(name, arguments, options, content, lineno,
                        content_offset, block_text, state, state_machine):
-    """"""
-    class_value = nodes.make_id(arguments[0])
-    if class_value:
-        pending = nodes.pending(misc.ClassAttribute,
-                                {'class': class_value, 'directive': name},
-                                block_text)
-        state_machine.document.note_pending(pending)
-        return [pending]
-    else:
+    """
+    Set a "class" attribute on the directive content or the next element.
+    When applied to the next element, a "pending" element is inserted, and a
+    transform does the work later.
+    """
+    try:
+        class_value = directives.class_option(arguments[0])
+    except ValueError:
         error = state_machine.reporter.error(
             'Invalid class attribute value for "%s" directive: "%s".'
             % (name, arguments[0]),
             nodes.literal_block(block_text, block_text), line=lineno)
         return [error]
+    node_list = []
+    if content:
+        container = nodes.Element()
+        state.nested_parse(content, content_offset, container)
+        for node in container:
+            node['classes'].extend(class_value)
+        node_list.extend(container.children)
+    else:
+        pending = nodes.pending(misc.ClassAttribute,
+                                {'class': class_value, 'directive': name},
+                                block_text)
+        state_machine.document.note_pending(pending)
+        node_list.append(pending)
+    return node_list
 
-class_directive.arguments = (1, 0, 0)
+class_directive.arguments = (1, 0, 1)
 class_directive.content = 1
 
 role_arg_pat = re.compile(r'(%s)\s*(\(\s*(%s)\s*\)\s*)?$'
@@ -276,6 +345,49 @@ def role(name, arguments, options, content, lineno,
 
 role.content = 1
 
+def default_role(name, arguments, options, content, lineno,
+                 content_offset, block_text, state, state_machine):
+    """Set the default interpreted text role."""
+    if not arguments:
+        if roles._roles.has_key(''):
+            # restore the "default" default role
+            del roles._roles['']
+        return []
+    role_name = arguments[0]
+    role, messages = roles.role(
+        role_name, state_machine.language, lineno, state.reporter)
+    if role is None:
+        error = state.reporter.error(
+            'Unknown interpreted text role "%s".' % role_name,
+            nodes.literal_block(block_text, block_text), line=lineno)
+        return messages + [error]
+    roles._roles[''] = role
+    # @@@ should this be local to the document, not the parser?
+    return messages
+
+default_role.arguments = (0, 1, 0)
+
+def title(name, arguments, options, content, lineno,
+          content_offset, block_text, state, state_machine):
+    state_machine.document['title'] = arguments[0]
+    return []
+
+title.arguments = (1, 0, 1)
+
+def date(name, arguments, options, content, lineno,
+         content_offset, block_text, state, state_machine):
+    if not isinstance(state, states.SubstitutionDef):
+        error = state_machine.reporter.error(
+            'Invalid context: the "%s" directive can only be used within a '
+            'substitution definition.' % (name),
+            nodes.literal_block(block_text, block_text), line=lineno)
+        return [error]
+    format = '\n'.join(content) or '%Y-%m-%d'
+    text = time.strftime(format)
+    return [nodes.Text(text)]
+
+date.content = 1
+
 def directive_test_function(name, arguments, options, content, lineno,
                             content_offset, block_text, state, state_machine):
     """This directive is useful only for testing purposes."""
